#ifndef dm_filter_H
#define dm_filter_H

/* Filtered DNA sequences */
/* Last edited on 2008-04-15 10:15:36 by stolfi */

#define dm_filter_H_COPYRIGHT \
  "Copyright  2005  by the State University of Campinas (UNICAMP)" \
  " and the Federal Fluminense University (UFF)"

#include <dm_basic.h>
#include <dm_sample.h>
#include <dm_datum.h>
#include <dm_seq.h>

#include <msm_rung.h>
#include <msm_pairing.h>
#include <msm_cand.h>

#include <vec.h>
#include <pswr.h>

/* This interface defines tools for filtering sequences 
  of multi-channel samples ({dm_datum_t}s).
  
  In a sequence of order zero, there is one datum for each codon, and
  the densities are computed for that single codon (either +1 or -1).
  A sequence of order {r > 0} is obtained by smoothing the sequence of
  order 0 with a Gaussian filter of appropriate width, and sampling
  the result with step 2 starting at 0.5. */

int dm_filter_output_num_sample_datums(int nold, bool_t circ, int nwtb);
  /* Number of samples in the sequence that will be returned by {dm-seq_filter},
    given that the input sequence has {nold} samples, circularity {circ},
    and the filter table has {nwtb} entries.
    
    If {circ} is true, the result is {nold/2} samples, 
    otherwise it is {(nold - wtbp->ne + 1)/2} samples */

dm_seq_t dm_filter_seq(dm_seq_t *oldp, double_vec_t *wtbp, char *wcmt, int nsub);
  /* Filters and resamples a numeric sequence at double step size.
    More precisely, the sample datums {newp[0..nnew-1]} of the new
    sequence are related to the sample datums {oldp[0..nold-1]} of the 
    old one by the formula  
    
      {newp[i] = SUM{ oldp[2*i+j] * wtbp[j] : j in 0..wtbp->ne-1}}
      
    Let {nold = dm_seq_num_sample_datums(oldp)}. If {oldp->circ} is
    true, then {nnew} is {(nold+1)/2} and the index index {2*i+j} is
    taken modulo {nold}. If {oldp->circ} is false, then {nnew} is
    {(nold-wtbp->ne+1)/2}, so that the index index {2*i+j} is is
    restricted to {0..nold-1}.
    
    The new sequence will have the same attributes {id,name,circ} as
    {oldp}, its {level} will be {oldp.sd.level+1}, and its subsampling
    factor will be {nsub}. For each channel [k], the new scale factor
    {sfac.f[k]} will be the root-mean-square value of the samples in
    channel {k} of the filtered sequecne. The string {wcmt}, if not
    NULL, is assumed to be the name or description of the weight table
    {wtb}, and is appended to the comment {oldp->cmt} to make the new
    string's comment. */

/* MULTISCALE FILTERING OF BIO-SEQUENCES */

void dm_filter_seq_multi
  ( dm_seq_t *s, 
    int maxLevel, 
    double_vec_t *wtb0, 
    char *wname0,
    double_vec_t *wtb1, 
    char *wname1,
    dm_seq_t sr[],
    int nsub
  );
  /* Places into {sr[0..maxLevel]} various filtered versions of the
    sequence {s}. Namely {sr[0]} is a storage-independent
    copy of {s}, and {sr[r]} is the filtered and down-sampled
    version of {sr[r-1]}, for {r} in {1..maxLevel}.  
    
    If a sequence {sr[r]} has {n} sample datums, then sequence {sr[r+1]}
    will have at most {(n+1)/2} sample datums, where {w} is the number of
    sample datums in the filter table. Therefore, the sequence {sr[r]} may
    be empty for {r} beyond a certain value.
    
    The table {wtb0} is used to go from {sr[0]} to {sr[1]}, and {wtb1}
    is used for the other scales; while the comment string {wname0} or
    {wname1}, respectively, is to the {cmt} field of each new
    sequence. */

/* AUXILIARY FUNCTIONS */

void dm_double_vec_filter(double_vec_t *oldp, double_vec_t *wtbp, double_vec_t *newp);
  /* Computes the convolution of the sample sequence {oldp} with the
    weight sequence {wtbp}, displaced by two steps at a time, and
    stores the result in {newp}. More precisely, sets {newp[i]} to 
    
      {SUM{ oldp[2*i+j] * wtbp[j] : j in 0..wtbp->ne-1}}
      
    where the index {2*i+j} is taken modulo {oldp->ne}. */

#endif
