#ifndef float_image_H
#define float_image_H

/* Multichannel images with floating-point samples. */
/* Last edited on 2008-05-24 23:57:20 by stolfi */ 

#include <stdio.h>
#include <bool.h>
#include <indexing.h>

#define INF INFINITY

typedef struct float_image_t 
  { ix_size_t sz[3]; /* Image sizes: sz[0] = channels, sz[1] = columns, sz[2] = rows. */
    ix_step_t st[3]; /* Indexing step along each axis. */
    ix_pos_t bp;     /* Position of sample {[0,0,0]}. */
    float *sample;   /* Linearized sample vector. */
  } float_image_t; 
  /* The sample value for channel {c}, column {x}, and row {y} 
    of the image is stored in {sample[bp + c*st[0] + x*st[1] + y*st[2]]}. */

#define float_image_max_size 65356
  /* Maximum dimension along any axis (2^16), for sanity checks. */

/* ALLOCATION */

float_image_t *float_image_new (int NC, int NX, int NY);
  /* Allocates a new multichannel image with {NC} channels,
    {NX} columns, and {NY} rows. */

float_image_t *float_image_copy (float_image_t *A);
  /* Creates a new float image with the same size and number
    of channels as {A}, and copies into its samples the samples of {A}. */

void float_image_free(float_image_t *A);
  /* Discards the pixel array *and the header* of the image {A}. */

/* ACESSING INDIVIDUAL SAMPLES */

float float_image_get_sample(float_image_t *A, int c, int x, int y);
  /* Returns the sample in channel {c}, column {x}, row {y} of image {A}. */

float *float_image_get_sample_address(float_image_t *A, int c, int x, int y);
  /* Returns the address of the sample in channel {c}, column {x}, row {y} 
    of image {A}. */

void float_image_set_sample(float_image_t *A, int c, int x, int y, float v);
  /* Stores {v} into the sample in channel {c}, column {x}, row {y} of image {A}. */

/* ACESSING WHOLE PIXELS */

void float_image_get_pixel(float_image_t *A, int x, int y, float v[]);
  /* Copies into {v[c]} the sample in channel {c}, column {x} and row
    {y} of image {A}, for {c=0..NC-1}; where {NC} is the number of
    channels of {A}. */

void float_image_set_pixel(float_image_t *A, int x, int y, float v[]);
  /* Stores the sample value {v[c]}, into channel {c}, column {x} and
    row {y} of image {A}, for {c=0..NC-1}; where {NC} is the number
    of channels of {A}. */
    
void float_image_fill_pixel(float_image_t *A, int x, int y, float v);
  /* Stores the sample value {v} into channel {c}, column {x} and row
    {y} of image {A}, for {c=0..NC-1}; where {NC} is the number of
    channels of {A}. */

/* IMAGE INTERPOLATION

  For the procedures in this section, the pixel on column {ix} and 
  row {iy} is assumed to be centered at point {(ix+0.5,iy+0.5)}. */

float float_image_interpolate_samples(float_image_t *A, int c, double x, double y);
  /* Returns the value of channel {c} of image {A} at the point
    {(x,y)}, computed by bilinear interpolation from nearby
    samples. 
    
    More precisely, if {x} and {y} are of the form {ix + fx} and {iy +
    fy}, with integer {ix,iy} and {fx,fy} in [-0.5 _ +0.5], then
    returns the weighted average of the four pixels in columns {ix-1}
    and {ix}, with weights {0.5-fx} and {0.5+fx}, and rows {iy-1} and
    {iy}, with weights {0.5-fy} and {0.5+fy}.
    
    For this formula, pixels outside the array bounds are assumed to
    be filled by duplicating the extremal rows and columns of
    {A}. */

void float_image_interpolate_pixels(float_image_t *A, double x, double y, float v[]);
  /* For {c=0..NC-1}, stores into {v[c]} the value of chanel {c} of
    image {A} at the point {(x,y)}, computed by bilinear
    interpolation from nearby samples; where {NC} is the number of
    channels of {A}. */

/* ACESSING WHOLE ROWS OF SAMPLES */

void float_image_get_sample_row(float_image_t *A, int c, int y, float v[]);
  /* Copies into {v[x]} the sample of channel {c}, column {x}, row {y}
    of image {A}, for {x=0..NX-1}; where {NX} is the number of
    columns of {A}. */

void float_image_set_sample_row(float_image_t *A, int c, int y, float v[]);
  /* Stores the sample value {v[x]} into channel {c}, column {x}, row {y} of
    image {A}, for {x=0..NX-1}; where {NX} is the number of columns
    of {A}. */
    
void float_image_fill_sample_row(float_image_t *A, int c, int y, float v);
  /* Stores the sample value {v} into channel {c}, column {x}, row {y}
    of image {A}, for {x=0..NX-1}; where {NX} is the number of
    columns of {A}. */

/* ACESSING WHOLE ROWS OF PIXELS */

void float_image_get_pixel_row(float_image_t *A, int y, float v[]);
  /* Copies into {v[c + x*NC]} the sample of channel {c}, column {x},
    row {y} of image {A}, for {c=0..NC-1} and {x=0..NX-1}; where
    {NC} and {NX} are the number of channels and the number of columns
    of {A}. */

void float_image_set_pixel_row(float_image_t *A, int y, float v[]);
  /* Stores the sample value {v[c + x*NC]} into channel {c}, column
    {x}, row {y} of image {A}, for {c=0..NC-1} and {x=0..NX-1};
    where {NC} and {NX} are the number of channels and the number of
    columns of {A}. */
    
void float_image_fill_pixel_row(float_image_t *A, int y, float v);
  /* Stores the sample value {v} into channel {c}, column {x}, row {y}
    of image {A}, for {c=0..NC-1} and {x=0..NX-1}; where {NC} and {NX}
    are the number of channels and the number of columns of {A}. */

/* ACCESSING WHOLE CHANNELS */

void float_image_fill_channel(float_image_t *A, int c, float v);
  /* Stores {v} into all samples in channel {c} of image {A}. */

/* ACCESSING THE WHOLE IMAGE */

void float_image_fill(float_image_t *A, float v);
  /* Stores {v} into all samples of image {A}. */

/* SAMPLE ENCODING AND RANGE */

void float_image_apply_gamma(float_image_t *A, int c, double gamma, double bias);
  /* Applies to channel {c} of image {A} the power correction with exponent {gamma}
    and offset {bias}. See {sample_conv_gamma} for details. */ 

void float_image_rescale_samples(float_image_t *A, int c, float a0, float a1, float z0, float z1);
  /* Applies to every sample of channel {c} of {A} the affine
    rescaling that takes the interval {[a0 _ a1]} to {[z0 _ z1]}. The
    samples are not clipped to either interval.
    
    The rescaling is defined as {v = z0 + (v-a0)*((z1-z0)/(a1-a0))}
    where {v} is the sample's value.
    
    There is no requirement that {a0<a1} or {z0<z1}. However, if {a0==a1}, 
    the rescaling may produce samples that are {INF} or {NAN}. */

void float_image_compute_sample_avg_dev(float_image_t *A, int c, double *avg, double *dev);
  /* Computes the average {*avg} and standard deviation {*dev} of all
    samples in channel {c} of image {A}. 
    
    Ignores samples that are {INF} or {NAN}. The deviation uses the
    unbiased variance estimator formula (with {N-1} rather than {N} in
    the denominator). If {c} is not a valid channel index, assumes
    that all samples in that channel are all zero. If there are no
    valid samples, both parameters are set to zero. If there is only
    one valid sample, the average is that sample, and the deviation is
    set to zero. */

void float_image_update_sample_range(float_image_t *A, int c, float *vMin, float *vMax);
  /* Expands the range {[*vmin _ *vmax]} so that it contains all
    finite sample values in channel {c} of image {A}. 
    
    Ignores samples that are {INF} or {NAN}. If {c} is not a valid
    channel index, assumes that all samples in that channel are all
    zero. If there are no valid samples, does not modify {*vMax}
    and {*vMin}. */

/* GET/CHECK SIZE */
  
void float_image_get_size(float_image_t *A, int *NC, int *NX, int *NY);
  /* Stores the number of channels, columns, and rows of {A} in
    {*NC}, {*NX}, and {*NY}, respectively. */
  
void float_image_check_size(float_image_t *A, int NC, int NX, int NY);
  /* Bombs out if {A} does not have exactly {NC} channels, {NX}
    columns, and {NY} rows. */

/* INPUT/OUTPUT */

void float_image_write(FILE *wr, float_image_t *A);
  /* Writes the image {A} to {wr}, in ASCII floating point format.
    The file will start with a standard header line 
    "begin float_image_t (version {DATE})", 
    followed by three lines with "NC = {NUM}", "NX = {NUM}", "NY =
    {NUM}". Then follow the pixels, one per line with "{X} {Y}
    {SAMPLE}...". A blank line is inserted between two consecutive
    rows. The file ends with a footer "end float_image_t". */

float_image_t *float_image_read(FILE *rd);
  /* Reads an image from {rd}, in ASCII floating point format
    generated by {float_image_write}. */

#endif
