/* Coordinate systems for images. */
/* Last edited on 2007-09-06 18:26:59 by stolfi */

#ifndef image_coords_H
#define image_coords_H

#include <bool.h>
#include <r2.h>
#include <r3x3.h>
#include <hr2.h>
#include <argparser.h>

hr2_pmap_t imgc_coord_sys_map
  ( bool_t hrev, 
    bool_t vrev, 
    bool_t center, 
    r2_t *org, 
    int cols, 
    int rows
  );
  /* Returns a projective map {M} that converts point coordinates from
    the {float_image.h} native coordinate system (FS) to the user's
    coordinate system (US).
    
    The FS has the X axis pointing to the right, the Y axis pointing
    up, and the origin at the bottom left corner of the image's
    domain.
    
    The US has the X axis pointing to the right if {hrev} is FALSE,
    and to the left if {hrev} is TRUE; the Y axis pointin up if {vrev}
    is FALSE, and down if {vrev} is true. The US origin is located at
    the center of the image's domain if {center} is TRUE, or at the
    position {org} if {center} is FALSE.  In the second case, each
    coordinate of {org} is measured from the edge of the image's
    domain that has lowest coordinate (that is, the left edge if
    {hrev} is FALSE, the right edge if {hrev} is true; the bottom edge
    if {vrev} is FALSE, the top edge if {vrev} is TRUE).
    
    In any case, all coordinates are measured in pixels. The image's
    domain is assumed to be a rectangle of width {cols} and height
    {rows}. This is equivalent to assuming that a pixel with indices
    {[ix,iy]} is a square, displaced {ix} units to the right and {iy}
    units up from the domain's lower left corner. */

/* COMMAND LINE PARSING */

/* General help about image coordinate systems */

#define imgc_axes_INFO \
  "Points in image domains" \
  " are expressed in terms of a coordinate system" \
  " specific to each image, input or output.  In both systems," \
  " the directions of the axes are determined by" \
  " the \"-haxis\" and \"-vaxis\" command line arguments.  These" \
  " arguments also affect the default origin of the coordinate" \
  " system.  In the default coordinate systems, the origin lies" \
  " at the low corner of the domain (which depends on the axis directions)."
  
#define imgc_pixel_centers_INFO \
  "The unit of length is the pixel size; in the default coordinate systems," \
  " pixel *corners* have integer coordinates, and pixel *centers* have half-integer coordinates."
  
/* Horizontal axis direction */

void imgc_parse_haxis(argparser_t *pp, bool_t *hleft);
  /*  Parses the "-haxis" command line option, according to
    {imgc_parse_haxis_HELP}. Sets {*hleft} to TRUE if "-haxis left" is
    present, to FALSE if "-haxis right" is present. If the option is
    not present, leaves {*hleft} unchanged. */

#define imgc_parse_haxis_HELP \
  "[ -haxis { right | left } ]"

#define imgc_parse_haxis_HELP_INFO \
  "  -haxis { right | left }\n" \
  "    This optional argument specifies the direction of the" \
  " horizontal axis of image coordinate systems.  It also defines the" \
  " default horizontal position of the origin: the left edge" \
  " of the image if \"right\", or the right edge" \
  " if \"left\"."
  
#define imgc_parse_haxis_pbm_default_INFO \
  "The default is \"-haxis right\" as in most image processing tools."
  
#define imgc_parse_haxis_math_default_INFO \
  "The default is \"-haxis right\" as in mathematical tradition."
  
/* Vertical axis direction */

void imgc_parse_vaxis(argparser_t *pp, bool_t *vdown);
  /*  Parses the "-vaxis" command line option, according to
    {imgc_parse_vaxis_HELP}. Sets {*vdown} to TRUE if the "-vaxis
    down" is present, to FALSE if "-vaxis up" is present. If the
    option is not present, leaves {*vdown} unchanged. */

#define imgc_parse_vaxis_HELP_INFO \
  "  -vaxis { down | up }\n" \
  "    This optional argument specifies the direction of" \
  " the vertical axis for both images.  It also defines the" \
  " default vertical position of the origin: the top edge of the" \
  " image if \"down\", or the bottom edge if \"up\"."
  
#define imgc_parse_vaxis_pbm_default_INFO \
  "The default is \"-yaxis down\" as in most image processing tools." \

#define imgc_parse_vaxis_math_default_INFO \
  "The default is \"-yaxis up\", in the mathematical tradition." \

#define imgc_parse_vaxis_HELP \
  "[ -vaxis { down | up } ]"
  
/* Input image origin */

void imgc_parse_icenter_iorg(argparser_t *pp, bool_t *icenter, r2_t *iorg);
  /* Parses the mutually exclusive arguments "-icenter" and "-iorg",
    that define the coordinate origin for input images,
    according to the syntax specs {imgc_parse_icenter_iorg_HELP}. If
    "-icenter"is preent, sets {*icenter} to TRUE and leaves {*iorg}
    unchanged. If "-iorg" is present, sets {*icenter} to FALSE, and sets
    {*iorg} to the given point. If neither is present, leaves {*icenter}
    and {*iorg} unchanged. */

#define imgc_parse_icenter_iorg_HELP \
  "[ -icenter | -iorg {CX_IN} {CY_IN} ] \\\n" \

#define imgc_input_origin_INFO \
  "The default coordinate system origin for input images" \
  " can be overriden by the \"-iorg\" or \"-icenter\" argument."  

#define imgc_parse_icenter_iorg_HELP_INFO \
  "  -icenter\n" \
  "  -iorg {CX_IN} {CY_IN}\n" \
  "    These mutually exclusive optional arguments specify" \
  " the origin of the coordinate system for input" \
  " images.  The \"-icenter\" option sets the origin" \
  " at the center of the image domain.  The \"-iorg\" option sets" \
  " the origin at the point {(CX_IN,CY_IN)} relative to the" \
  " default origin.  If this argument is not specified, the program" \
  " assumes \"-iorg 0 0\" (which means the default origin)."

/* Output image origin */

void imgc_parse_ocenter_oorg(argparser_t *pp, bool_t *ocenter, r2_t *oorg);
  /* Analogous to {imgc_parse_center_org}, but applying to output
    images, according to {imgc_parse_ocenter_oorg_HELP_INFO}. 
    Looks for the keywords "-ocenter" and "-oorg". */

#define imgc_parse_ocenter_oorg_HELP \
  "[ -ocenter | -oorg {CX_OUT} {CY_OUT} ] \\\n" \

#define imgc_output_origin_INFO \
  "The default coordinate system origin for output images" \
  " can be overriden by the \"-oorg\" or \"-ocenter\" argument."

#define imgc_parse_ocenter_oorg_HELP_INFO \
  "  -ocenter\n" \
  "  -oorg {CX_OUT} {CY_OUT}\n" \
  "    These mutually exclusive optional arguments specify" \
  " the origin of the coordinate system for output" \
  " images.  The \"-ocenter\" option sets the origin" \
  " at the center of the image domain.  The \"-oorg\" option sets" \
  " the origin at the point {(CX_OUT,CY_OUT)} relative to the" \
  " default origin.  If this argument is not specified, the program" \
  " assumes \"-oorg 0 0\" (which means the default origin).\n"

/* Output image size: */

void imgc_parse_osize(argparser_t *pp, int *ocols, int *orows, int max_size);
  /* Parses the output image size option "-osize", according to {imgc_parse_osize_HELP}. If
    "-osize" is present, sets {*ocols} and {*orows} to the next two
    arguments. Otherwise leaves {*ocols} and {*orows} unchanged.
    Fails with error if either dimension exceeds {max_size}. */
    
#define imgc_parse_osize_HELP \
  "[ -osize {NX} {NY} ]"

#define imgc_parse_osize_HELP_INFO \
  "  -osize {NX} {NY}\n" \
  "    Specifies the horizontal and vertical size of the output image."

#endif
