/* spmat.h -- sparse matrics (two-dimensional arrays) of things */
/* Last edited on 2008-03-29 10:31:41 by stolfi */

/* A /sparse matrix/ is a two-dimensional matrix that contains a large
  proportion of /trivial elements/ --- elements that are equal to
  zero, or to some other application-specific value.
  
  This module defines a /compressed representation/ for sparse
  matrices, namely a list of the element values together with their
  original row and column indices. Any element not on this list is assumed to be
  trivial. The representation is handled through a /descriptor/, a
  record containing the nominal dimensions (row and column counts) of the matrix,
  the count of elements stored in the list, and a pointer to an array
  of (row,col,value) triplets.
  
  This interface also defines efficient and convenient procedures for
  extending and trimming the list of entries, as needed. Each
  expansion doubles the size of the entry list, so the total cost of
  multiple expansions is small and proportional to the size of the
  final list. */

#ifndef spmat_H
#define spmat_H

#include <stdlib.h>
#include <bool.h>
#include <ref.h>
#include <stdint.h>

/* TYPED SPARSE MATRIX DESCRIPTORS
  
  To define a new sparse matrix type called, e.g., {my_spmat_t}, with
  elements of type {my_elem_t}, write
  
    { spmat_typedef(my_spmat_t,my_spmat,my_elem_t); }
    
  in the proper ".h" file, and 
  
    { spmat_typeimpl(my_spmat_t,my_spmat,my_elem_t); }
    
  in the corresponding ".c" file.

  The three parameters specify the matrix type name, the operation
  name prefix, and the element type. These three names can be chosen
  independently; however, it is recommended that they be of the form
  {XXX_spmat_t}, {XXX_spmat}, and {XXX_t}, respectively, for some
  {XXX}.
  
  These macros will declare the types
  {my_spmat_entry_t} (a (row,col,value) triplet) and {my_spmat_t} 
  (a sparse matrix descriptor), and the four associated procedures:
  
    { my_spmat_new
      my_spmat_expand
      my_spmat_trim
      my_spmat_unpack
      my_spmat_pack
      my_spmat_desc }
      
  which are described below.
    
  MATRIX ENTRY
  
  A variable {e} of type {my_spmat_entry_t} represents an explicit
  element of the sparse matrix. It contains the indices {e.row} and
  {e.col} of the element, and its value {e.val}.

  DESCRIPTOR CONTENTS
  
  A variable {M} of type {my_spmat_t} is a descriptor to a matrix of
  {my_elem_t} elements, whose nominal size is {nr} rows by {nc}
  columns, but having only {M.nen} explicit elements. The field {M.en}
  points to a heap area with space for {M.nen} explicit entries of type
  {my_spmat_entry_t}, namely {M.en[0..M.nen-1]}.

  ALLOCATION
  
  The call {my_spmat_new(nr, nc, nen)} allocates a new sparse matrix
  with nominal size {nr} by {nc} and space for {nen} explicit elements,
  and returns its descriptor.

  STORAGE EXPANSION
  
  The call {my_spmat_expand(&M,pos)} makes sure that the entry
  {M.en[pos]} exists. If necessary, it reallocates the area
  referenced by {M.en}, copies the old entries into it, and updates
  {M.nen} to the new size (which will be strictly greater than
  {pos}). The size {M.nen} is doubled at each reallocation, to ensure
  total {O(N)} time for adding {N} entries.

  STORAGE TRIMMING
  
  The call {my_spmat_trim(&M,size)} reallocates {M.en} if necessary so
  that it has exactly {size} explicit entries. It preserves the entries
  {M.en[0..size-1]} and sets {M.nen = size}.
  
  PACKING AND UNPACKING
  
  
  
  DESCRIPTOR ASSEMBLY
  
  The call {my_spmat_desc(nc, nr, nen, en)} assembles a sparse matrix
  descriptor from given row count {nr}, column count {nc}, explicit
  element count {nen}, and the address {en} of the first explicit
  entry. The client must ensure that the variables {en[0..nen-1]}
  actually exist. The procedures {my_spmat_expand} and {my_spmat_trim}
  can be applied to the resulting descriptor only if the address {en}
  was obtained through {malloc}.
  
*/

#define spmat_typedef(SPMATTYPE,SPMATNAME,ELEMTYPE) \
typedef struct SPMATNAME##_entry_t { uint32_t row, col; ELEMTYPE val; } SPMATNAME##_entry_t; \
\
typedef struct SPMATTYPE { uint32_t nr, nc; uint32_t nen; SPMATNAME##_entry_t *en; } SPMATTYPE; \
\
SPMATTYPE SPMATNAME##_new(uint32_t nr, uint32_t nc, uint32_t nen); \
\
void SPMATNAME##_expand(SPMATTYPE *nv, int pos); \
\
void SPMATNAME##_trim(SPMATTYPE *nv, uint32_t nen); \
\
SPMATTYPE SPMATNAME##_desc(uint32_t nr, uint32_t nc, SPMATNAME##_entry_t *en, uint32_t nen)

#define spmat_typeimpl(SPMATTYPE,SPMATNAME,ELEMTYPE) \
SPMATTYPE SPMATNAME##_new(uint32_t nr, uint32_t nc, uint32_t nen) \
  { spmat_t M = spmat_new(nr, nc, nen, sizeof(SPMATNAME##_entry_t));            \
    return (SPMATTYPE){M.nr, M.nc, M.nen, (ELEMTYPE *)M.en};              \
  } \
\
void SPMATNAME##_expand(SPMATTYPE *nv, int pos) \
  { if (pos >= nv->nen) \
      { spmat_expand((spmat_t *)nv, pos, sizeof(SPMATNAME##_entry_t)); } \
  } \
\
void SPMATNAME##_trim(SPMATTYPE *nv, uint32_t nen) \
  { spmat_trim((spmat_t *)nv, nen, sizeof(SPMATNAME##_entry_t)); } \
\
SPMATTYPE SPMATNAME##_desc(uint32_t nr, uint32_t nc, SPMATNAME##_entry_t *en, uint32_t nen) \
{ return (SPMATTYPE){nr, nc, nen, en}; }                                  \
\
extern void SPMATNAME##_bOgUs /* To eat the semicolon. */

/* GENERIC SPARSE MATRIX DESCRIPTORS */

typedef struct spmat_t /* General sparse matrix descriptor. */
  { uint32_t nr,   /* Number of rows. */
    uint32_t nc,   /* Number of columns. */
    uint32_t nen;   /* Number of explicit entries. */
    void *en;      /* Pointer to the first explicit entry. */
  } spmat_t;
  /* An {spmat_t} {M} describes a uni-dimensional array of {M.nen} elements
    of the same size and type, stored in consecutive memory locations,
    starting at address {M.en}. */

/* GENERIC SPARSE MATRIX FUNCTIONS 
  
  The functions in this section may be called directly by clients,
  but their main purpose is to implement the 
  functions {my_spmat_new}, {my_spmat_expand}, and {my_spmat_trim}
  defined by {spmat_typedef} and {spmat_typeimpl}. */

spmat_t spmat_new(uint32_t nr, uint32_t nc, int nen, size_t elsz);
  /* Allocates a new sparse matrix with {nen} elements of size {elsz}.
    Bombs out if there is no space for the request. 
    If {nen == 0}, the result has {en == NULL}. */

void spmat_expand(spmat_t *M, uint32_t pos, size_t elsz);
  /* Makes sure that element {M->en[pos]} exists, reallocating and
    copying the array {*M->en} if {pos >= M->nen}. If that happens,
    the new {M->nen} will be about twice as big as the old one. */

void spmat_trim(spmat_t *M, uint32_t nen, size_t elsz);
  /* Makes sure that {M.nen == nen}, reallocating and copying 
    the  array {*M->en} if {M.nen != nen}. If {nen == 0}, the 
    result will have {en == NULL}. */

/* SUB-SPARSE MATRIX DESCRIPTORS */

typedef struct spmat_sub_t /* Regular sub-sequence of a {spmat_t}. */
  { uint32_t nen;   /* Number of elements in sub-sparse matrix. */
    void *en;  /* Pointer to element zero of sub-sparse matrix. */
    int step;  /* Increment between consecutive elements (in elements). */
  } spmat_sub_t;

spmat_sub_t spmat_as_sub(spmat_t *M);
  /* Returns a sub-sparse matrix descriptor for the entire sparse matrix
    {M} (with {step == 1}). */
    
spmat_sub_t spmat_sub(spmat_sub_t *vs, uint32_t start, int step, uint32_t nen, size_t elsz);
  /* Returns a descriptor {r} for the specified subsequence of elements
    of {vs}, namely {r[i] = vs[start + i*step]}, for {i = 0..r.nen-1}.
    The {step} cannot be zero.
    
    The result size {r.nen} is the largest integer in {0..nen} such
    that all those elements exist. In particular, if {start} is not in
    {0..vs.nen-1}, then {m} is zero. */
    
/* SOME USEFUL TYPED SPARSE MATRICES */

spmat_typedef(int_spmat_t,    int_spmat,    int);
  /* Sparse Matrices of {int}s. */

spmat_typedef(uint_spmat_t,   uint_spmat,   uint32_t);
  /* Sparse Matrices of {uint32_t}s. */

spmat_typedef(float_spmat_t,  float_spmat,  float);
  /* Sparse Matrices of {float}s */

spmat_typedef(double_spmat_t, double_spmat, double);
  /* Sparse Matrices of {double}s. */

spmat_typedef(bool_spmat_t,   bool_spmat,   bool_t);
  /* Sparse Matrices of {bool_t}s. */

spmat_typedef(char_spmat_t,   char_spmat,   char);
  /* Sparse Matrices of {char}s. */

spmat_typedef(ubyte_spmat_t,   ubyte_spmat, uint8_t);
  /* Sparse Matrices of {unsigned char}s. */

spmat_typedef(sbyte_spmat_t,   sbyte_spmat, int8_t);
  /* Sparse Matrices of {signed char}s. */

spmat_typedef(string_spmat_t, string_spmat, char*);
  /* Sparse Matrices of strings ({char*}s). */

spmat_typedef(ref_spmat_t, ref_spmat, ref_t);
  /* Sparse Matrices of arbitrary addresses. */

#endif
