/* See ulist.h */
/* Last edited on 2008-02-25 09:16:16 by stolfi */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>

#include <vec.h>
#include <bool.h>
#include <ref.h>
#include <jsrandom.h>
#include <jsmath.h>
#include <affirm.h>

#include <ulist.h>
  
#define ulist_DEBUG 1
  /* Define this to 1 to cause the printout of debugging info
    and performance statistics. */

#define ulist_PARANOIA 1
  /* Define this to 1 to call {ulist_verify} after main operations. */

vec_typedef(ulist_hash_val_vec_t, ulist_hash_val_vec, ulist_hash_val_t);
vec_typedef(ulist_index_vec_t, ulist_index_vec, ulist_index_t);

struct ulist_t {
  /* The item list: */
  ulist_item_vec_t it;        /* The item storage slots. */
  ulist_count_t ct;           /* The current count of items in the {ulist_t}. */
  /* The hashing tables: */
  ulist_hash_val_vec_t hx;    /* Maps item indices to (displaced) hash values. */
  ulist_index_vec_t ix;       /* Maps (displaced) hash values to item indices. */
  /* The methods: */
  ulist_eq_func_t *eq;        /* The item equivalence predicate. */
  ulist_hash_func_t *hash;    /* The item hash function. */
  #if (ulist_DEBUG)
  /* Debugging data: */
  uint64_t ct_add;            /* Count of {ulist_insert_last} operations. */
  uint64_t ct_del;            /* Count of {ulist_delete_last} operations. */
  uint64_t ct_ind;            /* Count of {ulist_index_of} operations. */
  uint64_t ct_rsz;            /* Count of {ulist_resize} operations. */
  uint64_t ct_prb;            /* Count of probes in hash table search. */
  #endif /* (ulist_DEBUG) */
};

/* INTERNAL NOTATION
  
  In the following comments, `{~a}' means `any {ulist_item_t} value {b} such
  that {eq(a,b)} is TRUE'. 
  
  We also write 
  
    {ct} instead of {ct(S)}. 
    
    {sz} instead of {sz(S)} (that is, {S->it.ne}).
    
    {nh} instead of {nh(S)}, the size of the hash table. 
    
    {it[i]} for the variable {S->it.e[i]}, for any {i} in {0..ct-1}.
    
    {hx[i]} for the variable {S->hx.e[i]}, for any {i} in {0..ct-1}.
    
    {ix[k]} for the variable {S->ix[imod(k,nh)]}, for any integer {k}.
  
  Note that {ix} is conceptually a circular array with period {nh},
  while {it} and {hx} are ordinary (non-circular) arrays.
  
  Also, by {i..j(%n)} we denote all integers {r} in the range {i..i+d}
  where {d = imod(j-i,n)}; namely, the {d+1} consecutive integers that
  start with {i} and continue *forward* until the first integer
  congruent to {j} modulo {n} --- even if {i < j}. In particular, if
  {j==i} (or {j == i+k*n} for any {k}) then {d == 0} and {i..j(%n)} is
  just one integer {i == j}. Otherwise {i..j(%n)} has between 2 and
  {n-1} integers.
  
  The notation {ix[i..j]} is then defined as {ix[r]} for all {r} in {i..j(%nh)}.

  STRUCTURE INVARIANTS:

    I0: {ct == S->ct} and {sz == S->it.ne} and {nh = S->ix.ne}.
    
    I1: {ct <= sz <= ulist_item_count_MAX}.
    
    I2: {S->xh.ne == sz}.
    
    I3 (list): The item of {S} with index {i} is {it[i]}, for {i} in {0..ct-1}.
    
    I4 (bilink): for any {i} in {0..ct-1}, {(hx[i] < nh) && (ix[hx[i]] == i)}. 
    
    I5 (seq-search): for any {i} in {0..ct-1}, {occ(r)} is true for any
       {r} in {h..k(%nh)}, where {h = hash(it[i],nh)} and {k = hx[i]}.

    I6 (distinctness): for any {i,j} in {0..ct-1}, {eq(it[i],it[j])} implies {i==j}.
    
    I7. Either {sz == nh == 0}, or {sz < nh <= ulist_hash_size_MAX}
  
  Invariant I4 means that an entry {ix[k]} is associated to an item
  {it[i]} iff {ix[k] == i} and {hx[i] == k}. Entries of {ix[0..nh-1]}
  that pass this test are `occupied', all other are `vacant'.
  Therefore we define the predicate {occ(k)}, for {k} in {0..nh-1}, as
  {(ix[k] < ct) && (k == hx[ix[k]])}; and {vac(k)} as {! occ(k)}.

  Invariants {I4} and {I5} imply that, if {~a} is in {S}, its (unique) index
  {i} will be found in the {ix} table by searching consecutive entries
  {ix[h]}, {ix[h+1]}, {ix[h+2]}, ..., starting at its hash value {h =
  hash(a,nh)}, before finding a vacant entry.
 
 OCCUPANCY RATIO AND PERFORMANCE

  Some {ulist_t} operations call the {ulist_locate} function, that 
  finds the entry in the {ix} table for an element {a}, given the 
  value of {hash(a,nh)}.  The expected cost of the {locate} 
  function generally depends on the /relative occupancy/ of the 
  {ix} table, that is, {ro == ct/nh}.
  
  If {ro} is well below 1, the expected cost of {locate} is about {K1
  + K2*ro}. As {ro} approaches 1, the expected cost increases to
  {K2*ct}.
  
  The cost of deleting an item is about {K1 + K2*ro^2} if {ro} is
  small, and can be very high ( perhaps {K1 + K2*ct*log(ct)}? ) as {ro}
  approaches {1}. Therefore, to preserve the efficiency of those
  operations, we keep {nh} approximately equal to {2*sz}. */

ulist_hash_size_t ulist_choose_hash_size(ulist_count_t n);
/* Chooses a suitable hash table size for a list with capacity {n}. */

/* PROCEDURES THAT DO NOT DEPEND ON HASHING */

vec_typeimpl(ulist_item_vec_t, ulist_item_vec, ulist_item_t);
vec_typeimpl(ulist_hash_val_vec_t, ulist_hash_val_vec, ulist_hash_val_t);
vec_typeimpl(ulist_index_vec_t, ulist_index_vec, ulist_index_t);

ulist_t *ulist_new(ulist_count_t n)
  { demand(n <= ulist_item_count_MAX, "requested capacity is too large");
    ulist_t *S = (ulist_t *)notnull(malloc(sizeof(ulist_t)), "out of mem");
    S->it = ulist_item_vec_new(n);
    S->ct = 0;
    S->hx = ulist_hash_val_vec_new(n);
    ulist_hash_size_t nh = ulist_choose_hash_size(n); 
    S->ix = ulist_index_vec_new(nh);
    S->eq = &ulist_uint64_eq;
    S->hash = &ulist_uint64_hash;
    #if (ulist_DEBUG)
    ulist_stats_clear(S);
    #endif /* (ulist_DEBUG) */
    return S;
  }
  
ulist_hash_size_t ulist_choose_hash_size(ulist_count_t n)  
  { if (n == 0) { /* No need for a hash table: */ return 0; }
    double dnh = 2.0*((double)n); /* Ideal size of hash table. */
    demand(dnh <= (double)ulist_hash_size_MAX, "hash table would be too big");
    ulist_hash_size_t nh = (ulist_hash_size_t)floor(dnh);
    /* If {n} is large enough, we had better round it up to avoid small divisors: */
    uint32_t pmax = 13; /* Largest prime we plan to avoid. */
    uint32_t m = 1; /* Product of all small primes we intend to avoid. */
    uint32_t r = 0; /* Desired remainder modulo {m}. */
    uint32_t p = 1; /* Largest prime in {m}, or 1: */
    if (ulist_DEBUG) { fprintf(stderr, "  n = %u  ideal nh = %d", n, nh); }
    while (TRUE)
      { /* Find the next prime {p}: */
        do { p++; } while (gcd(m, p) != 1); 
        if (p > pmax) { /* No more primes to avoid: */ break; }
        if (m > 3*nh/2/p) { /* Avoiding {p} would require too many entries: */ break; }
        /* Adjust {r} so that it is about {m/2} and {gcd(r,m*p) = 1}: */
        unsigned int q = (int)(floor(0.61803398874989484820*p));
        r = q*m + r;
        /* Include {p} in {m}: */
        m = p*m;
        if (ulist_DEBUG) { fprintf(stderr, " [k*%d + %d]", m, r); }
      }
    assert(m < ulist_hash_size_MAX); /* Paranoia... */
    /* Round {nh} down to a multiple {nhs} of {m}: */
    ulist_hash_size_t nhs = (nh / m) * m;
    /* Ensure that {nhs + r} is at least {nh}: */
    if (r < nh - nhs) { r += m; }
    /* Ensure that {nhs + r} is not too big: */
    while (r > ulist_hash_size_MAX - nhs)
      { assert(nhs >= m);
        nhs -= m;
      }
    /* Phew: */
    nh = nhs + r;
    if (ulist_DEBUG) { fprintf(stderr, "  nh = %u\n", nh); }
    return nh;
  }

void ulist_free(ulist_t *S)
  { free(S->it.e); 
    free(S->ix.e); 
    free(S->hx.e); 
    free(S); 
  }

void ulist_clear(ulist_t *S)
  { S->ct = 0; }

ulist_eq_func_t *ulist_get_eq(ulist_t *S)
  { return S->eq; }

void ulist_install_eq(ulist_t *S, ulist_eq_func_t *eq)
  { if (eq == NULL) { /* Provide default: */ eq = &ulist_uint64_eq; }
    if (eq != S->eq)
      { demand((S->ct <= 1), "cannot change {eq} of non-trivial list"); 
        S->eq = eq;
      }
  }

ulist_hash_func_t *ulist_get_hash(ulist_t *S)
  { return S->hash; }

void ulist_install_hash(ulist_t *S, ulist_hash_func_t *hash)
  { if (hash == NULL) { /* Provide default: */ hash = &ulist_uint64_hash; }
    if (hash != S->hash)
      { demand((S->ct == 0), "cannot change {hash} of non-empty list"); 
        S->hash = hash;
      }
  }

ulist_count_t ulist_count(ulist_t *S)
  { return S->ct; }

ulist_count_t ulist_capacity(ulist_t *S)
  { return S->it.ne; }

ulist_hash_size_t ulist_hash_size(ulist_t *S)
  { return S->ix.ne; }

ulist_item_vec_t ulist_items(ulist_t *S)
  { ulist_count_t ct = S->ct;
    /* Allocate the result vector: */
    ulist_item_vec_t v = ulist_item_vec_new(ct);
    /* Copy the items to {v}: */
    ulist_count_t i;
    ulist_item_t *p; /* Pointer that scans {S->it.e}. */
    ulist_item_t *q; /* Pointer that scans {v.e}. */
    for (i = 0, p = S->it.e, q = v.e; i < ct; i++, p++, q++) { (*q) = (*p); }
    return v;
  }

ulist_item_t ulist_item_at(ulist_t *S, ulist_index_t i)
  { ulist_count_t ct = S->ct;
    demand((i >= 0) && (i < ct), "invalid item index");
    return S->it.e[i];
  }

/* PROCEDURES THAT DEPEND ON HASHING */

void ulist_locate(ulist_t *S, ulist_item_t a, ulist_index_t *ip, ulist_hash_val_t *kp);
  /* If {~a} is in {S}, the procedure returns in {*ip} the index {i} 
    of {~a}, and in {*kp} the position {k} of that index in {ix[0..nh-1]}
    (that is, returns {i} and {k} such that {eq(it[i],a)} and {ix[k] == i}).
    
    If {~a} is not in {S}, returns in {*ip} the current count {ct}, and
    in {*kp} the position {k} of the vacant entry {ix[k]} where the index
    of {a} would have to be stored if it were appended to the list. 
    
    Fails if {ct >= nh}. */

void ulist_locate(ulist_t *S, ulist_item_t a, ulist_index_t *ip, ulist_hash_val_t *kp)
  { /* Get the hash table size {nh} and the item count {ct}: */
    ulist_hash_size_t nh = S->ix.ne;
    ulist_count_t ct = S->ct;
    demand(nh > 0, "should not have called locate");
    assert(ct < nh); /* By the invariants. */
    /* Get the hash value {h} of {a}: */
    ulist_hash_val_t h = S->hash(a, nh);
    demand(h < nh, "bad hash value");
    /* Sequential serch in {ix} starting at {ix[h]}: */
    ulist_hash_val_t k = h;
    while (TRUE) { 
      /* Get the presumed index {i}: */
      #if (ulist_DEBUG)
      S->ct_prb++;
      #endif /* (ulist_DEBUG) */
      ulist_index_t i = S->ix.e[k]; 
      /* fprintf (stderr, "  %d %d\n", k, i); */
      /* Check its validity: */
      if ((i >= ct) || (S->hx.e[i] != k)) 
        { /* Entry {ix[k]} is vacant, so the search is over (not found): */
          assert((k >= 0) && (k < nh)); 
          (*kp) = k; (*ip) = ct; 
          return;
        }
      /* Entry {ix[k]} is occupied: */
      ulist_item_t b = S->it.e[i]; 
      if (S->eq(a,b))
        { /* The search is over (found): */
          (*kp) = k; (*ip) = i;
          return;
        }
      /* Advance {k} to the next entry: */
      k++; if (k == nh) { k = 0; }
      /* Since {nh > ct}, we will stop before {k} returns to {h}. */ 
    }
  }

ulist_index_t ulist_index_of(ulist_t *S, ulist_item_t a)
  { 
    #if (ulist_DEBUG)
    S->ct_ind++;
    #endif /* (ulist_DEBUG) */
    /* Get the item count {ct}: */
    ulist_count_t ct = S->ct;
    if (ct == 0) { /* List is empty: */ return 0; } 
    /* Locate the index {i} and hash table slot {k} of {~a}: */
    ulist_index_t i;
    ulist_hash_val_t k;
    ulist_locate(S, a, &i, &k);
    return i;
  }

bool_t ulist_has(ulist_t *S, ulist_item_t a)
  { return (ulist_index_of(S,a) != S->ct); }

ulist_index_t ulist_insert_last(ulist_t *S, ulist_item_t a)
  { 
    #if (ulist_DEBUG)
    S->ct_add++;
    #endif /* (ulist_DEBUG) */
    /* Get the list count {ct} and capacity {sz}: */
    ulist_count_t ct = S->ct;
    ulist_count_t sz = S->it.ne;
    /* Can we add another element? */
    demand(ct < sz, "list is full");
    /* !!! implement auto-resize if full. !!! */
    /* Locate the index {i} and hash table slot {k} of {~a}: */
    ulist_index_t i;
    ulist_hash_val_t k;
    ulist_locate(S, a, &i, &k);
    if ((i == 749) && (k == 2180)) { (void)ulist_verify(S, TRUE); }
    demand(i == ct, "element is already in table");
    /* Insert item: */
    S->it.e[i] = a;
    S->hx.e[i] = k;
    S->ix.e[k] = i;
    S->ct++;
    #if (ulist_PARANOIA)
    (void)ulist_verify(S, TRUE);
    #endif /* (ulist_PARANOIA) */
    return i;
  }

ulist_item_t ulist_delete_last(ulist_t *S)
  { 
    #if (ulist_DEBUG)
    S->ct_del++;
    #endif /* (ulist_DEBUG) */
    /* Get the list count {ct}: */
    ulist_count_t ct = S->ct;
    /* We cannot delete from an empty list: */
    demand(ct > 0, "list is empty");
    /* Get the item {a} to be deleted: */
    ulist_item_t a = S->it.e[ct-1]; 
    /* Locate the index {i} and hash table slot {k} of {~a}: */
    ulist_index_t i;
    ulist_hash_val_t k;
    ulist_locate(S, a, &i, &k);
    assert(i == ct-1);
    /* Grab the item {b} to be deleted: */
    /* bool_t debug = FALSE; */
    /* bool_t debug = ((i == 749) && (k == 0)); */
    bool_t debug = ((S->ix.ne == 3711) && (a == 5833));
    if (debug) { (void)ulist_verify(S, TRUE); }
    /* bool_t debug = ((sz == 1000) && (ct == 751)); */
    /* Remove {a} from {S}: */
    S->ct--;
    /* Now the items in occupied entries that follow {ix[k]} may need reinserting: */
    if (debug) { fprintf(stderr, "-- start actual deletion --\n"); }
    if (debug) { fprintf(stderr, "  vacated ix[%u]\n", k); }
    if (debug) { (void)ulist_verify(S, FALSE); }
    /* Get the hash table size {nh}: */
    ulist_hash_size_t nh = S->ix.ne;
    if (debug) { fprintf(stderr, "  nh = %u\n", nh); }
    unsigned int r = (k + 1) % nh; /* Scans entries after {S[k]}. */
    unsigned int d = 0; /* Always equal to {imod(r - (k+1), nh)}. */ 
    while(TRUE)
      { /* Now entries {ix[k+1..r]\ix[r]}} are OK and occupied. */
        if (debug) { fprintf(stderr, "  k = %u  r = %u  d = %u\n", k, r, d); }
        ulist_index_t j = S->ix.e[r];
        if ((j >= ct) || (S->hx.e[j] != r))
          { if (debug) { fprintf(stderr, "  found vacant ix[%u]\n", r); }
            /* If our logic is correct, there are no more inconsistencies. */
            break;
          }
        /* Slot {r} is occupied by item {it[j]}. */
        if (debug) { fprintf(stderr, "  ix[%u] occupied by %u", r, j); }
        /* At this point the slot {ix[r]} is distinct from {ix[k]}. */
        /* Check whether {it[j]} needs to be reinserted: */
        ulist_hash_val_t hj = S->hash(S->it.e[j], nh);
        demand (hj < nh, "bad hash function");
        unsigned int dhj = imod(hj - (k+1), nh);
        if (debug) { fprintf(stderr, "  it[%u] hashes to %u  dhj = %u  d = %u", j, hj, dhj, d); }
        if (dhj > d) 
          { /* Item {it[j]} is not OK, must remove and reinsert it. */
            if (debug) { fprintf(stderr, "  presumed BAD\n"); }
            /* Store the index {j} into {ix[k]}: */
            S->ix.e[k] = j; S->hx.e[j] = k; 
            if (debug) { fprintf(stderr, "  filled ix[%u] with ix[%u] = %u\n", k, r, j); }
            /* Now the hot hole is {ix[r]}: */
            if (debug) { fprintf(stderr, "  ix[%u] is now vacant\n", r); }
            k = r; d = 0; 
            if (debug) { (void)ulist_verify(S, FALSE); }
          }
        else
          { if (debug) { fprintf(stderr, "  presumed OK\n"); } }
        /* Go on to the next slot of {ix[]}: */
        d++; r++; if (r == nh) { r = 0; }
      }
    #if (ulist_PARANOIA)
    (void)ulist_verify(S, TRUE);
    #endif /* (ulist_PARANOIA) */
    if (debug) { fprintf(stderr, "-- end actual deletion --\n"); }
    if (debug) { (void)ulist_verify(S, TRUE); }
    /* !!! implement auto-resize if less than 1/2 full. !!! */
    return a;
  }

ulist_item_t ulist_set(ulist_t *S, ulist_index_t i, ulist_item_t a)
  { /* Get the list count {ct}: */
    ulist_count_t ct = S->ct;
    /* Check index validity: */
    demand((i >= 0) && (i < ct), "invalid item index");
    /* Swap item number {i} with last item: */
    ulist_swap(S, i, ct-1);
    /* Delete the item number {i}, now in last position, and save it in {b}: */
    ulist_item_t b = ulist_delete_last(S); 
    /* Append the new item: */
    ulist_insert_last(S, a);
    /* Swap it back to position {i}: */
    ulist_swap(S, i, ct-1);
    /* We are done: */
    return b;
  }

void ulist_swap(ulist_t *S, ulist_index_t i, ulist_index_t j)
  { /* Get the list count {ct}: */
    ulist_count_t ct = S->ct;
    /* Validate the indices: */
    demand(i < ct, "bad index i");
    demand(j < ct, "bad index j");
    /* Trivial case: */
    if (i == j) { return; }
    /* Swap the entries of {it[]}: */
    { ulist_item_t t = S->it.e[i]; S->it.e[i] = S->it.e[j]; S->it.e[j] = t; }
    /* Swap the entries of {hx[]}: */
    { ulist_hash_val_t t = S->hx.e[i]; S->hx.e[i] = S->hx.e[j]; S->hx.e[j] = t; }
    /* Fix the links in {ix[]}: */
    S->ix.e[S->hx.e[i]] = i;
    S->ix.e[S->hx.e[j]] = j;
  }

void ulist_resize(ulist_t *S, int n, ulist_hash_func_t *hash)
  { 
    #if (ulist_DEBUG)
    S->ct_rsz++;
    #endif /* (ulist_DEBUG) */
    if (hash == NULL) { /* Provide default: */ hash = &ulist_uint64_hash; }
    ulist_count_t sz = S->it.ne;
    if ((sz != n) || (S->hash != hash))
      { /* Must actually do a resize: */
        /* Grab the item vector {it} of {S}: */
        ulist_item_vec_t old_it = S->it;
        /* Replace {S->it} by a new item vector with {n} slots: */
        S->it = ulist_item_vec_new(n);
        /* Make {S} empty and change its hash function: */
        S->ct = 0;
        S->hash = hash;
        /* Items are non-eq already, so can use {==} instead of {eq} for reinsertion: */
        ulist_eq_func_t *eq = S->eq;
        S->eq = ulist_uint64_eq;
        /* Insert all items of {old_it} back into {S}. */
        ulist_index_t i;
        ulist_item_t *p; /* Pointer that scans the item slots. */
        for (i = 0, p = old_it.e; i < sz; i++, p++)
          { ulist_item_t a = (*p);
            ulist_insert_last(S, a);
          }
        /* Restore the original equivalence: */
        S->eq = eq;
        /* Reclaim the old vector: */
        free(old_it.e);
      }
  }

/* DEBUGGING */

bool_t ulist_verify(ulist_t *S, bool_t die)
  { ulist_count_t ct = S->ct;         /* Current item count. */
    ulist_count_t sz = S->it.ne;     /* Capacity */
    if (sz != S->hx.ne) { fail_test(die, "inconsistent sizes of it,hx"); }
    if (ct > sz) { fail_test(die, "size greater than capacity"); }
    ulist_hash_size_t nh = S->ix.ne; /* Hash table size. */
    if (sz == 0)
      { if (nh != 0) { fail_test(die, "zero capacity list with hash table"); } }
    else
      { if (nh <= sz) { fail_test(die, "hash table too small"); } }
    /* Check the structural invariants: */
    int i;
    for (i = 0; i < ct; i++)
      { ulist_item_t a = S->it.e[i];
        ulist_hash_val_t k = S->hx.e[i];
        if (k > nh) { fail_test(die, "invalid hash position hx[i]"); }
        if (S->ix.e[k] != i) { fail_test(die, "ix[hx[i]] is not i"); }
        /* Check the seqsearch invariant: */
        ulist_index_t j;
        ulist_hash_val_t kj;
        ulist_locate(S, a, &j, &kj);
        if (j != i) 
          { fprintf(stderr, "%s:%d: ** seq search invariant violated for it[%u]\n", __FILE__, __LINE__, i); 
            fprintf(stderr, "  sz = %u ct = %u nh = %u\n", sz, ct, nh); 
            ulist_hash_val_t ha = S->hash(a,nh);
            fprintf(stderr, "  hash(it[%u]) = %u  locate = it[%u],ix[%u]", i, ha, j,kj); 
            if (j < ct) 
              { ulist_item_t b = S->it.e[j];
                ulist_hash_val_t hb = S->hash(b,nh);
                fprintf(stderr, "  hash(it[%u]) = %u\n", j, hb);
              }
            else
              { fprintf(stderr, "  (not in list)\n"); }
            fail_test(die, "aborting");
          }
      }
    return TRUE;
  }

void ulist_stats_print(ulist_t *S)
  {
    #if (ulist_DEBUG)
    fprintf(stderr, "........................................................................\n");
    fprintf(stderr, "operation counts:\n");
    /* Basic counts: */
    fprintf(stderr, "  {ulist_insert_last} = %12lld\n", S->ct_add);
    fprintf(stderr, "  {ulist_delete_last} = %12lld\n", S->ct_del);
    fprintf(stderr, "  {ulist_index_of} = %12lld\n", S->ct_ind);
    fprintf(stderr, "  {ulist_resize} = %12lld\n", S->ct_rsz);
    fprintf(stderr, "  probes in hash table = %12lld\n", S->ct_prb);
    /* Probes per operation (including the resize overhead): */
    double ppo = ((double)S->ct_prb)/((double)S->ct_add + S->ct_del + S->ct_ind); 
    fprintf(stderr, "  probes per operation = %8.2f\n", ppo);
    fprintf(stderr, "........................................................................\n");
    #else 
    fprintf(stderr, "%s compiled without statistics support\n", __FILE__);
    #endif /* (ulist_DEBUG) */
  }

void ulist_stats_clear(ulist_t *S)
  {
    #if (ulist_DEBUG)
    S->ct_add = 0;
    S->ct_del = 0;
    S->ct_ind = 0;
    S->ct_prb = 0;
    S->ct_rsz = 0;
    #endif /* (ulist_DEBUG) */
  }

/* DEFAULT EQ/HASH FUNCTIONS */

ulist_hash_val_t ulist_uint64_hash(ulist_item_t a, ulist_hash_size_t n)
  { demand(n > 0, "cannot hash to empty range");
    /* Let's hope that this works: */
    uint64_t ua = (uint64_t) a;
    uint64_t ub = ua & 8191ULL;
    uint64_t ur = (4615ULL * (ua >> 14)) + (417ULL * (ua >> 9)) + (ub * ub * 471703ULL);
    ulist_hash_val_t u = (ur % n);
    return u;
  }

bool_t ulist_uint64_eq(ulist_item_t a, ulist_item_t b)
  { 
    return (a == b);
  }
  
