/* See pswr.h */
/* Last edited on 2008-05-25 03:26:13 by stolfi */

#include <pswr.h>
#include <pswr_def.h>
#include <pswr_aux.h>
#include <pswr_vis.h>

#include <affirm.h>
#include <bool.h>
#include <jsfile.h>
#include <jstime.h>
#include <jsstring.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

/* STREAM STATE */

bool_t pswr_within_canvas(PSStream *ps)
  { return ps->curSlot != INT_MAX; }

bool_t pswr_next_slot_available(PSStream *ps)
  { return (ps->curSlot != INT_MAX) && (ps->curSlot+1 < ps->hPicCount*ps->vPicCount); }

bool_t pswr_canvas_dirty(PSStream *ps)
  { return (ps->curSlot != INT_MAX) && (ps->curSlot >= 0); }

void pswr_begin_canvas(PSStream *ps, const char *pageName)
  { 
    affirm(! pswr_within_canvas(ps), "bad state");
    ps->curCanv++;
    affirm(ps->curCanv > 0, "curCanv bug");
    /* Provide a default for {pageName}, if NULL or empty: */
    bool_t pageNameOK = (pageName != NULL) && (*pageName != '\000');
    char pageNameBuf[7];
    if (! pageNameOK)
      { snprintf(pageNameBuf, 7, "%06d", ps->curCanv);
        pageName = &(pageNameBuf[0]);
      }
    if (ps->eps)
      { if(ps->file == NULL)
          { /* Open new file: */
            int nameSz = strlen(ps->name) + strlen(pageName) + 6;
            char fileName[nameSz];
            snprintf(fileName, nameSz, "%s-%s.eps", ps->name, pageName);
            ps->file = open_write(fileName, TRUE);
          }
        pswr_begin_encapsulated_figure(ps);
      }
    else
      { affirm(ps->file != NULL, "file bug");
        pswr_begin_document_page(ps, pageName);
      }
    /* Write canvas preamble commands to file: */
    double *fc = ps->fillColor;
    fc[0] = fc[1] = fc[2] = 0.500;
    pswr_write_canvas_graphics_setup_cmds(ps->file, fc);
    /* Mark writer state as {within_canvas}: */
    ps->curSlot = -1;
    /* Set the plot window to the full canvas: */
    double hSize = ps->hCanvasSize;
    double vSize = ps->vCanvasSize;
    /* Set window in writer (note: clobbers {ps->curSlot}, resets cell grid): */
    pswr_set_window(ps, 0.0, hSize, 0.0, vSize, 0.0, hSize, 0.0, vSize);
    /* Ensure that next picture slot is the first one: */
    ps->curSlot = -1;
  }

void pswr_end_canvas(PSStream *ps)
  { 
    affirm(pswr_within_canvas(ps), "bad state");
    affirm(ps->curCanv >= 0, "curCanv bug");
    affirm(ps->file != NULL, "no file");
    if (ps->eps)
      { /* We have an open file with stuff in it: */
        pswr_end_encapsulated_figure(ps);
        fclose(ps->file); 
        ps->file = NULL;
      }
    else
      { pswr_end_document_page(ps); }
    /* Indicate {between_canvas} state: */
    ps->curSlot = INT_MAX;
  }

/* FILE INTIALIZATION AND FINALIZATION */

void pswr_begin_encapsulated_figure(PSStream *ps)
  { affirm(ps->eps, "bad file type");
    
    /* Initialize the font list with the default fonts: */
    pswr_initialize_font_list(&(ps->nFonts), &(ps->fonts));

    /* Initialize local state: */
    double *fc = ps->fillColor;
    fc[0] = fc[1] = fc[2] = 0.500;

    FILE *file = ps->file;
    char *date = today();
    pswr_write_eps_file_header(file, date, 0.0, ps->hCanvasSize, 0.0, ps->vCanvasSize);
    if(ps->verbose) { fprintf(stderr, "date = %s\n", date); } 
    pswr_write_canvas_header(file, "1", 1);
    fflush(file);
    free(date);
  }
  
void pswr_end_encapsulated_figure(PSStream *ps)
  { FILE *file = ps->file;
    affirm(ps->eps, "use of pswr_end_figure in non-EPS file");
    pswr_write_canvas_trailer(file, TRUE);
    pswr_write_eps_file_trailer(file, ps->nFonts, ps->fonts);
    fflush(file);
    pswr_free_font_list(&(ps->nFonts), &(ps->fonts));
  }

void pswr_begin_document(PSStream *ps)
  { affirm(! ps->eps, "bad file type");

    /* Initialize the font list with the default fonts: */
    pswr_initialize_font_list(&(ps->nFonts), &(ps->fonts));

    /* Initialize local state: */
    FILE *file = ps->file;
    char *date = today();
    if (ps->verbose) { fprintf(stderr, "date = %s\n", date); }
    pswr_write_ps_file_header(file, ps->paperSize, date);
    fflush(file);
    free(date);
  }

void pswr_begin_document_page(PSStream *ps, const char *pageName)
  { affirm(! ps->eps, "use of pswr_begin_document_page in EPS file");
    affirm(! pswr_within_canvas(ps), "page begin/end bug");
    
    FILE *file = ps->file;
    char *date = today();
    if (ps->verbose) { fprintf(stderr, "page %s date = %s\n", pageName, date); }
    pswr_write_canvas_header(file, pageName, ps->curCanv);
    pswr_write_date_and_page_show_cmds(file, date, pageName);
    free(date);
  }

void pswr_end_document_page(PSStream *ps)
  { FILE *file = ps->file;
    affirm(! ps->eps, "use of pswr_end_document_page in EPS file");
    affirm(pswr_within_canvas(ps), "page begin/end bug");
    pswr_write_canvas_trailer(file, FALSE);
  }

void pswr_end_document(PSStream *ps)
  { affirm(! ps->eps, "use of pswr_end_document in EPS file");
    affirm(! pswr_within_canvas(ps), "page begin/end bug");
    /* In case the client forgot to call pswr_end_page: */
    FILE *file = ps->file;
    pswr_write_ps_file_trailer(file, ps->nCanv, ps->nFonts, ps->fonts);
    fflush(file);
    pswr_free_font_list(&(ps->nFonts), &(ps->fonts));
  }

/* PAGE LAYOUT TOOLS */

void pswr_fix_slot_size
  ( double usableSize, 
    double picMargin,
    double capSize,
    int *count,
    double *picSize, 
    double *slotSize
  )
  { 
    if ((*picSize) <= 0)
      { if ((*count) <= 0)
          { (*slotSize) = usableSize;
            (*count) = 1;
          }
        else 
          { (*slotSize) = usableSize / (*count);
          }
        (*picSize) = (*slotSize) - 2*picMargin - capSize;
      }
    else
      { (*slotSize) = (*picSize) + 2*picMargin + capSize; }
      
    if ((*count) <= 0)
      { (*count) = (int)floor((usableSize + 0.1)/(*slotSize));
        if ((*count) <= 0) { (*count) = 1; }
      }
  }

/* INTERNAL WINDOW SETUP */

void pswr_save_window_data
  ( PSStream *ps,
    double xMin, double xMax,
    double yMin, double yMax,
    double hMin, double hMax,
    double vMin, double vMax
  )
  { /* Save the actual plot window: */
    ps->hMin = hMin; ps->hMax = hMax;
    ps->vMin = vMin; ps->vMax = vMax;
    /* Save the client scales: */
    ps->xMin = xMin; ps->xMax = xMax;
    ps->yMin = yMin; ps->yMax = yMax;
    /* Compute client-to-file scales: */
    ps->xScale = (hMax - hMin)/(xMax - xMin);
    ps->yScale = (vMax - vMin)/(yMax - yMin);
    /* Check for equal scales, allowing for some rounding errors: */
    { double scalesum = fabs(ps->xScale) + fabs(ps->yScale);
      double scalediff = fabs(ps->xScale - ps->yScale);
      affirm(scalediff/scalesum < 0.0001, "unequal scales");
    }
  }

void pswr_save_grid_data(PSStream *ps, int xn, int yn)
  { /* Save client grid size: */
    ps->xGridCount = xn; 
    ps->yGridCount = yn;
  }

/* DOCUMENT FONT LIST */

void pswr_initialize_font_list(int *nFontsP, char ***fontsP)
  {
    affirm((*fontsP) == NULL, "non-NULL font list");
    /* "Courier" is default, for page numbers etc. */
    int nFonts = 1;
    char **fonts = (char **)malloc(sizeof(char *));
    fonts[0] = txtcat("Courier", ""); 
    /* Update caller's variables: */
    (*nFontsP) = nFonts;
    (*fontsP) = fonts;
  }

void pswr_add_font(const char *font, int *nFontsP, char ***fontsP)
  {
    int nFonts = (*nFontsP);
    char **fonts = (*fontsP);
    /* Check whether font already is in the table: */
    int i;
    for (i = 0; i < nFonts; i++)
      { if (strcmp(fonts[i], font) == 0) { return; } }
    /* Ensure that there is space for a new entry: */
    if (((nFonts - 1) & nFonts) == 0)
      { /* Table is full ({nFonts} is a power of 2), reallocate it. */
        fonts = realloc(fonts, 2*nFonts*sizeof(char*)); 
        affirm(fonts != NULL, "out of mem"); 
      }
    /* Store a copy of the name: */
    fonts[nFonts] = txtcat(font, "");
    nFonts++;
    /* Update caller's variables: */
    (*nFontsP) = nFonts;
    (*fontsP) = fonts;
  }

void pswr_free_font_list(int *nFontsP, char ***fontsP)
  {
    int nFonts = (*nFontsP);
    char **fonts = (*fontsP);
    int i;
    for (i = 0; i < nFonts; i++) { free(fonts[i]); }
    free(fonts);
    (*fontsP) = NULL;
    (*nFontsP) = 0;
  }

/* INTERNAL FILE WRITING PROCEDURES */

void pswr_write_eps_file_header
  ( FILE *file, 
    const char *date,
    double hMin, double hMax,
    double vMin, double vMax
  )
  { fprintf(file, 
      "%%!PS-Adobe-2.0 EPSF-2.0\n"
    );
    fprintf(file, 
      "%%%%CreationDate: %s\n", date
    );
    fprintf(file, 
      "%%%%BoundingBox: %.0f %.0f %.0f %.0f\n",
      floor(hMin), floor(vMin), ceil(hMax), ceil(vMax)
    );
    fprintf(file, 
      "%%%%Pages: 1\n"
    );
    fprintf(file, 
      "%%%%DocumentFonts: (atend)\n"
    );
    fprintf(file, 
      "%%%%EndComments\n"
    );
    
    fprintf(file, 
      "/pswr$dict 6400 dict def \n"
      "pswr$dict begin\n"
    );
    pswr_write_proc_defs(file);
    fprintf(file, 
      "end\n"
    );

    fprintf(file, 
      "%%%%EndProlog\n"
    );
  }

void pswr_write_eps_file_trailer(FILE *file, int nFonts, char **fonts)
  { fprintf(file, "%%%%Trailer\n" );
    pswr_write_font_list(file, nFonts, fonts);
  }

void pswr_write_ps_file_header(FILE *file, const char *paperSize, const char *date)
  { fprintf(file, 
      "%%!PS-Adobe-2.0\n"
      "%%%%Pages: (atend)\n"
    );
    fprintf(file, 
      "%%%%DocumentFonts: (atend)\n"
    );
    fprintf(file, 
      "%%%%DocumentPaperSizes: %s\n", paperSize
    );
    fprintf(file, 
      "%%%%EndComments\n"
    );
    
    fprintf(file, 
      "/pswr$dict 6400 dict def \n"
      "pswr$dict begin\n"
    );
    pswr_write_proc_defs(file);
    fprintf(file, 
      "end\n"
    );

    fprintf(file, 
      "%%%%EndProlog\n"
    );
  }

void pswr_write_canvas_header(FILE *file, const char *pageName, int pageSeq)
  { if ((pageName != NULL) && (*pageName != '\000'))
      { fprintf(file, "%%%%Page: %d %d\n\n", pageSeq, pageSeq); }
    else
      { fprintf(file, "%%%%Page: %s %d\n\n", pageName, pageSeq); }

    fprintf(file, 
      "%% Save the standard coord system:\n"
      "/savedstate save def\n"
      "\n"
      "pswr$dict begin\n"
      "\n"
    );
    fflush(file);
  }

void pswr_write_canvas_trailer(FILE *file, int eps)
  { if (! eps) { fprintf(file, "showpage\n"); }
    fprintf(file, 
      "end %% pswr$dict\n"
      "\n"
      "savedstate restore\n"
      "%% Now we are back to the initial Postscript state system.\n"
      "\n"
    );
    fprintf(file, 
      "%%%%EndPage\n"
      "\n"
    );
    fflush(file);
  }
  
void pswr_write_ps_file_trailer(FILE *file, int nCanv, int nFonts, char **fonts)
  { fprintf(file, 
      "%%%%Trailer\n"
    );
    pswr_write_font_list(file, nFonts, fonts);
    fprintf(file, 
      "%%%%Pages: %d\n", nCanv
    );
  }
  
void pswr_write_font_list(FILE *file, int nFonts, char **fonts)
  { int i;
    fprintf(file, "%%%%DocumentFonts:");
    for (i = 0; i < nFonts; i++)
      { fprintf(file, " %s", fonts[i]); }
    fprintf(file, "\n");
  }

void pswr_write_date_and_page_show_cmds(FILE *file, const char *date, const char *page)
  { 
    if (date == NULL) { date = ""; }
    if (page == NULL) { page = ""; }
    char *fill = (strlen(page) == 0 ? "" : "  page ");
    char buf[strlen(date) + strlen(fill) + strlen(page) + 10]; /* 10 for safety... */
    sprintf(buf, "%s%s%s", date, fill, page); 
    fprintf(file, 
      "%% Print date and page:\n"
      "gsave\n"
      "  /Courier findfont 10 scalefont setfont\n"
      "  72 40 moveto\n"
    );
    fprintf(file, "  ");
    pswr_write_ps_string(file, buf, "\\267");
    fprintf(file, " show\n");
    fprintf(file, 
      "grestore\n"
      "\n"
    );
  }

void pswr_write_canvas_graphics_setup_cmds(FILE *file, double *fc)
  { fprintf(file, 
      "%% Round joints and caps:\n"
      "1 setlinecap 1 setlinejoin\n"
      "\n"
      "%% Black thin lines:\n"
      "0 setlinewidth 0 setgray [ ] 0 setdash\n"
      "\n"
    );
    pswr_write_fill_color_set_cmd(file, fc);

    /* Just in case the client forgets to set it up: */
    pswr_write_label_font_setup_cmds(file, "Courier", 8.0);    
  }
  
void pswr_write_window_setup_cmds
  ( FILE *file,
    double hMin, double hMax,
    double vMin, double vMax
  )
  { /* In spite of their names, the Postscript variables
      {xmin,xmax,ymin,ymax} are not client coordinates, 
      but actual plotting coordinates (in pt), namely 
      {hMin,hMax,vMin,vMax}. */
      
    fprintf(file, "/xmin %f def   %% min plottable x\n", hMin);
    fprintf(file, "/xmax %f def   %% max plottable x\n", hMax);

    fprintf(file, "/ymin %f def   %% min plottable y\n", vMin);
    fprintf(file, "/ymax %f def   %% max plottable y\n", vMax);

    fprintf(file, 
      "%% Set clipping path to boundary of plot area:\n"
      "initclip\n"
      "newpath\n"
      "  xmin ymin moveto\n"
      "  xmax ymin lineto\n"
      "  xmax ymax lineto\n"
      "  xmin ymax lineto\n"
      "  xmin ymin lineto\n"
      "clip\n"
      "\n"
    );

    /* Caption cursor. */
    
    fprintf(file, 
      "%% Left margin cursor for caption text:\n"
      "/dytext 10 pt mul def\n"
      "/xtext xmin def\n"
      "/ytext ymin def\n"
      "\n"
    );

    fprintf(file, 
      "%% Caption font setup:\n"
      "/captionfont\n"
      "  /Courier findfont\n"
      "  dytext scalefont\n"
      "def\n"
      "\n"
    );

    fflush(file);
  }

void pswr_write_grid_setup_cmds(FILE *file, int xn, int yn)
  { fprintf(file, "/xn %d def     %% grid cells along x axis\n", xn);
    fprintf(file, "/yn %d def     %% grid cells along y axis\n", yn);

    fprintf(file, "/xstep xmax xmin sub xn div def  %% x-size of grid cell\n");
    fprintf(file, "/ystep ymax ymin sub yn div def  %% y-size of grid cell\n");

    fflush(file);
  }

void pswr_write_label_font_setup_cmds(FILE *file, const char *font, double size)
  { fprintf(file, 
      "%% Label font setup:\n"
      "/labelfont\n"
      "  /%s findfont %.3f pt mul scalefont\n"
      "def\n" 
      "\n",
      font, size
    );
  }  

void pswr_write_color(FILE *file, double *clr)
  { fprintf(file, "  %5.3f %5.3f %5.3f", clr[0], clr[1], clr[2]); }

void pswr_write_fill_color_set_cmd(FILE *file, double *fc)
  { pswr_write_color(file, fc);
    fprintf(file, " sfc\n");
  }

void pswr_write_ps_string(FILE *file, const char *text, const char *newline)
  { const char *p;
    putc('(', file);
    for (p = text; *p != 0; p++)
      { if (*p == '\n')
          { fprintf(file, "%s", newline); }
        else if (*p == '(')
          { putc('\\', file); putc('(', file); }
        else if (*p == ')')
          { putc('\\', file); putc(')', file); }
        else if (*p == '\t')
          { putc(' ', file); putc(' ', file); }
        else if (*p == '\\')
          { putc('\\', file); putc('\\', file); }
        else if ((*p < ' ') || (*p > '~'))
          { fprintf(file, "\\%03o", *p); }
        else
          { putc(*p, file); }
      }
    fprintf(file, ")");
  }

void pswr_write_proc_defs(FILE *file)
  { 
    /* Global constants and variables: */
    
    fprintf(file, 
      "%% Units of measure:\n"
      "/pt 1.0 def\n"
      "/in pt 72.0 mul def \n"
      "/mm pt 72.0 25.4 div mul def\n"
      "\n"
    );
  
    fprintf(file, 
      "%% Segment draw operator:\n"
      "%%   /xa/ /ya/ /xb/ /yb/ segd --> \n"
      "/segd\n"
      "{ gsave\n"
      "    newpath\n"
      "    moveto\n"
      "    lineto\n"
      "    stroke\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Curve draw operator:\n"
      "%%   /xa/ /ya/  /xb/ /yb/  /xc/ /yc/  /xd/ /yd/ arcd --> \n"
      "/arcd\n"
      "{ gsave\n"
      "    newpath\n"
      "      8 -2 roll moveto curveto\n"
      "    stroke\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Draw an X-value grid line:\n"
      "%%   /x/ xgrd --> \n"
      "/xgrd\n"
      "{ gsave\n"
      "  newpath\n"
      "    dup ymin moveto\n"
      "    ymax lineto\n"
      "    stroke\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Draw an Y-value grid line:\n"
      "%%   /y/ ygrd --> \n"
      "/ygrd\n"
      "{ gsave\n"
      "  newpath\n"
      "    dup xmin exch moveto\n"
      "    xmax exch lineto\n"
      "    stroke\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Draw all grid lines:\n"
      "%%   gridlines --> \n"
      "/gridlines\n"
      "{ gsave\n"
      "    initclip\n"
      "    0 1 xn {\n"
      "      xstep mul xmin add xgrd\n"
      "    } for\n"
      "    0 1 yn {\n"
      "      ystep mul ymin add ygrd\n"
      "    } for\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file,
      "%% Draw a frame around the current plot window:\n"
      "%%   frame --> \n"
      "/wframe\n"
      "{ gsave\n"
      "    initclip\n"
      "    newpath\n"
      "    xmin ymin moveto\n"
      "    xmax ymin lineto\n"
      "    xmax ymax lineto\n"
      "    xmin ymax lineto\n"
      "    xmin ymin lineto\n"
      "    closepath stroke\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );
    
    /* Fill color: */
    
    fprintf(file, 
      "%% Set fill color operator:\n"
      "%%   /R/ /G/ /B/ sfc --> \n"
      "/sfc\n"
      "{ /fB exch def\n"
      "  /fG exch def\n"
      "  /fR exch def\n"
      "} bind def\n"
      "\n"
    );

    /* Figures: */
    
    fprintf(file, 
      "%% Fill and/or stroke operator:\n"
      "%%   /draw/ /fill/ fs --> \n"
      "/fs\n"
      "{ dup 2 index add\n"
      "  %% --- draw fill draw+fill\n"
      "  dup 0 eq\n"
      "  %% --- draw fill draw+fill neither\n"
      "    { pop pop pop newpath }\n"
      "    {\n"
      "      %% --- draw fill draw+fill\n"
      "      2 eq\n"
      "        { gsave fR fG fB setrgbcolor fill grestore stroke pop pop }\n"
      "        { 1 eq { fR fG fB setrgbcolor fill } if\n"
      "          1 eq { stroke } if\n"
      "        }\n"
      "      ifelse\n"
      "    }\n"
      "  ifelse\n"
      "} bind def\n"
      "\n"
    );
    
    fprintf(file, 
      "%% Even-odd fill and/or stroke operator:\n"
      "%%   /draw/ /fill/ eofs --> \n"
      "/eofs\n"
      "{ dup 2 index add\n"
      "  %% --- draw fill draw+fill\n"
      "  dup 0 eq\n"
      "  %% --- draw fill draw+fill neither\n"
      "    { pop pop pop newpath }\n"
      "    {\n"
      "      %% --- draw fill draw+fill\n"
      "      2 eq\n"
      "        { gsave fR fG fB setrgbcolor eofill grestore stroke pop pop }\n"
      "        { 1 eq { fR fG fB setrgbcolor eofill } if\n"
      "          1 eq { stroke } if\n"
      "        }\n"
      "      ifelse\n"
      "    }\n"
      "  ifelse\n"
      "} bind def\n"
      "\n"
    );
    
    fprintf(file, 
      "%% Triangle operator:\n"
      "%%   /draw/ /fill/ /xa/ /ya/ /xb/ /yb/ /xc/ /yc/ tri --> \n"
      "/tri\n"
      "{ gsave\n"
      "    newpath\n"
      "    moveto\n"
      "    lineto\n"
      "    lineto\n"
      "    closepath\n"
      "    fs\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );
    
    fprintf(file, 
      "%% Rectangle operator:\n"
      "%%   /draw/ /fill/ /xlo/ /xhi/ /ylo/ /yhi/ rec --> \n"
      "/rec\n"
      "{ gsave\n"
      "  newpath\n"
      "    3 index 2 index moveto\n"
      "    2 index 2 index lineto\n"
      "    2 index 1 index lineto\n"
      "    3 index 1 index lineto\n"
      "    pop pop pop pop\n"
      "    closepath\n"
      "    fs\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );
    
    fprintf(file, 
      "%% Polygon operator:\n"
      "%%   /draw/ /fill/ /x1/ /y1/ .. /xn/ /yn/ /n/ pol --> \n"
      "/pol\n"
      "{ gsave\n"
      "    newpath\n"
      "    3 1 roll moveto\n"
      "    1 sub 1 1  3 2 roll { pop lineto } for\n"
      "    closepath\n"
      "    fs\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Even-odd polygon operator:\n"
      "%%   /draw/ /fill/ /x1/ /y1/ .. /xn/ /yn/ /n/ eopol --> \n"
      "/eopol\n"
      "{ gsave\n"
      "    newpath\n"
      "    3 1 roll moveto\n"
      "    1 sub 1 1  3 2 roll { pop lineto } for\n"
      "    closepath\n"
      "    eofs\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Circle operator:\n"
      "%%   /draw/ /fill/ /x/ /y/ /rad/ cir --> \n"
      "/cir\n"
      "{ gsave\n"
      "    newpath\n"
      "    0 360 arc\n"
      "    closepath\n"
      "    fs\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );
    
    fprintf(file, 
      "%% Lune (lens) operator:\n"
      "%%   /draw/ /fill/ /xc/ /yc/ /rad/ /tilt/ lun --> \n"
      "/lun\n"
      "{ gsave\n"
      "    4 2 roll\n"
      "    %% --- draw fill rad tilt xc yc\n"
      "    translate\n"
      "    %% --- draw fill rad tilt\n"
      "    rotate\n"
      "    %% --- draw fill rad \n"
      "    newpath\n"
      "    dup\n"
      "    %% --- draw fill rad rad\n"
      "    dup neg 0\n"
      "    %% --- draw fill rad rad -rad 0\n"
      "    3 2 roll 2 mul\n"
      "    %% --- draw fill rad -rad 0  2*rad\n"
      "    -60 60 arc\n"
      "    %% --- draw fill rad\n"
      "    dup 0\n"
      "    %% --- draw fill rad rad 0\n"
      "    3 2 roll 2 mul\n"
      "    %% --- draw fill rad 0 2*rad\n"
      "    120 240 arc\n"
      "    closepath\n"
      "    fs\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Pie slice operator:\n"
      "%%   /draw/ /fill/ /xc/ /yc/ /rad/ /start/ /stop/ pie --> \n"
      "/pie\n"
      "{ gsave\n"
      "    newpath\n"
      "    4 index 4 index moveto arc\n"
      "    closepath\n"
      "    fs\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Grid cell operator:\n"
      "%%   /draw/ /fill/ /xi/ /yi/ cel --> \n"
      "/cel\n"
      "{ exch dup \n"
      "  %% --- draw fill yi, xi, xi\n"
      "  xstep mul xmin add exch 1 add xstep mul xmin add\n"
      "  %% --- draw fill yi, xlo, xhi\n"
      "  3 2 roll dup\n"
      "  %% --- draw fill xlo, xhi, yi, yi\n"
      "  ystep mul ymin add exch 1 add ystep mul ymin add\n"
      "  %% --- draw fill xlo, xhi, ylo, yhi\n"
      "  rec\n"
      "} bind def\n"
      "\n"
    );

    /* Labels and captions: */
    
    fprintf(file, 
      "%% Label printing operator:\n"
      "%%   /str/ /xa/ /ya/ /xc/ /yc/ lbsh --> \n"
      "/lbsh\n"
      "{ labelfont setfont\n"
      "  newpath moveto\n"
      "    %% --- str, xa, ya\n"
      "  gsave 2 index false charpath flattenpath pathbbox grestore\n"
      "    %% --- str, xa, ya, lox, loy, hix, hiy\n"
      "  3 index 3 index currentpoint \n"
      "    %% --- str, xa, ya, lox, loy, hix, hiy, lox, loy, cx, cy\n"
      "  exch 4 1 roll exch sub\n"
      "  3 1 roll sub exch\n"
      "    %% --- str, xa, ya, lox, loy, hix, hiy, cx-lox, cy-loy\n"
      "  rmoveto\n"
      "    %% --- str, xa, ya, lox, loy, hix, hiy\n"
      "  exch 4 1 roll exch sub \n"
      "  3 1 roll sub exch\n"
      "    %% --- str, xa, ya, dx, dy\n"
      "  exch 4 1 roll mul -1 mul\n"
      "  3 1 roll mul -1 mul exch\n"
      "    %% --- str, -dx*xa, -dy*ya\n"
      "  rmoveto\n"
      "    %% --- str\n"
      "  show\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Operator to move caption line cursor to new line:\n"
      "%%   nl --> \n"
      "/nl\n"
      "{ /ytext ytext dytext sub def\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Operator to print aligned caption line at CP without clipping:\n"
      "%%   /s/ /xa/ capsh --> \n"
      "/tmpstr 50 string def\n"
      "/capsh\n"
      "{ gsave\n"
      "    initclip\n"
      "    captionfont setfont\n"
      "    gsave\n"
      "      newpath 0 0 moveto\n"
      "      1 index false charpath flattenpath pathbbox\n"
      "    grestore\n"
      "      %% --- str, xa, lox, loy, hix, hiy\n"
      "      %% --- str, xa, lox, loy, hix, hiy\n"
      "    4 2 roll pop pop pop\n"
      "      %% --- str, xa, hix\n"
      "    xmax xmin sub\n"
      "      %% --- str, xa, hix, xmax-xmin\n"
      "    exch sub mul 0\n"
      "      %% --- str, xa*((xmax-xmin) - hix)), 0\n"
      "    newpath xtext ytext moveto\n"
      "      %% --- str, xa*((xmax-xmin) - hix)), 0\n"
      "    rmoveto\n"
      "      %% --- str\n"
      "    show\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fprintf(file, 
      "%% Operator to print string at CP without clipping:\n"
      "%%   /s/ shw --> \n"
      "/shw\n"
      "{ gsave\n"
      "    initclip\n"
      "    captionfont setfont show\n"
      "  grestore\n"
      "} bind def\n"
      "\n"
    );

    fflush(file);
  }
